from .qemu import qemu_provider
from .remote import remote_provider
from .request import request_machine

__all__ = ["qemu_provider", "remote_provider", "providers", "request_machine"]


providers = {
    "qemu": qemu_provider,
}



import requests
import urllib
import json

from .machines import MachineWrapper
from . import dictobj
from .users import User
from .request import request_machine, MachineRequest


def remote_provider(url=None, hostname=None, port=None):
    if url is None:
        if hostname is not None and port is not None:
            url = "http://{0}:{1}/".format(hostname, port)
        else:
            raise TypeError("Must provide either: url, or; hostname and port")
    return RemoteProvider(url)


class RemoteProvider(object):
    def __init__(self, base_url):
        self._api = RemoteApi(base_url)
        
    def start(self, *args, **kwargs):
        # TODO: remove duplication with QemuProvider.start
        if len(args) == 1 and not kwargs and isinstance(args[0], MachineRequest):
            request = args[0]
        else:
            request = request_machine(*(["peachtree"] + list(args)), **kwargs)
        response = self._api.start(request)
        return _create_machine(response, self._api)

    def find_running_machine(self, identifier):
        response = self._api.running_machine(identifier)
        if response is None:
            return None
        else:
            return _create_machine(response, self._api)
    
    def list_running_machines(self):
        machines = self._api.running_machines()
        return [_create_machine(machine, self._api) for machine in machines]
    
    def _url(self, path):
        return "{0}/{1}".format(self._base_url.rstrip("/"), path.lstrip("/"))
    
    def __enter__(self):
        return self
        
    def __exit__(self, *args):
        pass


def _create_machine(*args, **kwargs):
    return MachineWrapper(RemoteMachine(*args, **kwargs))


class RemoteMachine(object):
    def __init__(self, desc, api):
        self.identifier = desc["identifier"]
        self.name = desc["name"]
        self.image_name = desc["imageName"]
        self._hostname = desc["hostname"]
        self._users = [dictobj.dict_to_obj(user, User) for user in desc["users"]]
        self._api = api
    
    def hostname(self):
        return self._hostname
        
    def users(self):
        return self._users
    
    def is_running(self):
        return self._api.is_running(self.identifier)
    
    def public_port(self, guest_port):
        return self._api.public_port(self.identifier, guest_port)
    
    def restart(self):
        return self._api.restart(self.identifier)
    
    def destroy(self):
        self._api.destroy(self.identifier)
    
    def __repr__(self):
        return "RemoteMachine {0}".format(self.identifier)


class RemoteApi(object):
    _action_timeout = 60
    _info_timeout = 10
    
    def __init__(self, base_url):
        self._base_url = base_url
        
    def start(self, request):
        return self._action(
            "machines",
            data=dictobj.obj_to_dict(request),
        )
        
    def running_machine(self, identifier):
        return self._info(self._machine_path(identifier))
        
    def running_machines(self):
        return self._info("machines", data=None)
        
    def is_running(self, identifier):
        response = self._info(self._machine_path(identifier, "is-running"))
        return response["isRunning"]
        
    def public_port(self, identifier, port):
        return self._info(
            self._machine_path(identifier, "public-port"),
            data={"guest-port": port},
        )["port"]
        
    def restart(self, identifier):
        self._action(self._machine_path(identifier, "restart"))
        
    def destroy(self, identifier):
        self._action(self._machine_path(identifier, "destroy"))

    def _action(self, *args, **kwargs):
        return self._request(
            "POST", *args, timeout=self._action_timeout, **kwargs)
        
    def _info(self, *args, **kwargs):
        return self._request(
            "GET", *args, timeout=self._info_timeout, **kwargs)

    def _request(self, method, path, timeout, data=None):
        response = requests.request(
            method,
            self._url(path),
            data=json.dumps(data),
            headers={"Content-Type": "application/json"},
            timeout=timeout
        )
        if response.status_code not in [200, 404]:
            raise RuntimeError("Got response: {0}", response)
        return response.json()
        

    def _url(self, path):
        return "{0}/{1}".format(self._base_url.rstrip("/"), path.lstrip("/"))
        
    def _machine_path(self, identifier, extra=None):
        path = "machines/{0}".format(urllib.quote(identifier))
        if extra is None:
            return path
        else:
            return "{0}/{1}".format(path, extra)
        



import os
import tempfile
import errno
import json

import spur
import psutil

from . import dictobj


local_shell = spur.LocalShell()


def start(commands):
    run_dir = RunDirectory(tempfile.mkdtemp(prefix="process-set-"))
    process_set = ProcessSet(run_dir, {})
    process_set.start(commands)
    return process_set


def from_dir(run_dir):
    run_dir = RunDirectory(run_dir)
    names = run_dir.read_names()
    process_infos = map(run_dir.read_process_info, names)
        
    return ProcessSet(run_dir, dict(zip(names, process_infos)))


ProcessInfo = dictobj.data_class("ProcessInfo", ["pid", "start_time"])
    
    
class ProcessSet(object):
    def __init__(self, run_dir, processes):
        # TODO: should rename RunDirectory (process info persistence?)
        self._run_dir = run_dir
        self.run_dir = run_dir._run_dir
        self._processes = processes

    def start(self, commands):
        for name in commands.iterkeys():
            if name in self._processes:
                raise ValueError("Cannot start second process with name {0}".format(name))
        
        self._run_dir.append_names(commands.keys())
        
        def start_process((name, command_args)):
            output_file = self._run_dir.output_path(name)
            command = " ".join(map(_escape_sh, command_args))
            redirected_command = ["sh", "-c", "exec {0} > {1} 2>&1".format(command, output_file)]
            process = local_shell.spawn(redirected_command, store_pid=True)
            process_info = _process_info_for_pid(process.pid)
            self._run_dir.write_process_info(name, process_info)
                
            return (name, process_info)
            
        self._processes.update(map(start_process, commands.iteritems()))

    def all_running(self):
        return all(self._is_running_each_process())
    
    def any_running(self):
        return any(self._is_running_each_process())
    
    def all_output(self):
        names = sorted(self._processes.keys())
        return "".join(
            "{0}:\n{1}".format(name, _indent(self._output(name)))
            for name in names
        )
        
    def kill_all(self):
        for process_info in self._processes.itervalues():
            _kill(process_info)
            
    
    def _is_running_each_process(self):
        return map(_process_is_running, self._processes.itervalues())
    
    def _output(self, name):
        # TODO: remove duplication
        output_file = os.path.join(self.run_dir, "{0}.output".format(name))
        try:
            with open(output_file) as output:
                return output.read()
        except IOError as error:
            if error.errno == errno.ENOENT:
                return ""
            else:
                raise


def _indent(string):
    def indent_line(line):
        if line:
            return "    {0}\n".format(line)
        else:
            return line
        
    lines = string.split("\n")
    indented_lines = map(indent_line, lines)
    return "".join(indented_lines)


def _process_is_running(process_info):
    try:
        process = psutil.Process(process_info.pid)
    except psutil.NoSuchProcess:
        return False
    if process.status in [psutil.STATUS_DEAD, psutil.STATUS_ZOMBIE]:
        return False
    return _process_info_for_pid(process_info.pid) == process_info


def _kill(process_info):
    if _process_is_running(process_info):
        local_shell.run(["kill", str(process_info.pid)])

def _process_info_for_pid(pid):
    start_time = _process_start_time_from_pid(pid)
    return ProcessInfo(pid, start_time)
    

def _process_start_time_from_pid(pid):
    return psutil.Process(pid).create_time
    
    
def _escape_sh(value):
    return "'" + value.replace("'", "'\\''") + "'"


class RunDirectory(object):
    def __init__(self, run_dir):
        self._run_dir = run_dir
    
    def append_names(self, names):
        with open(self._names_path(), "a") as names_file:
            for name in names:
                names_file.write("{0}\n".format(name))
    
    def read_names(self):
        with open(self._names_path()) as names_file:
            return [line.strip() for line in names_file.readlines()]
    
    def write_process_info(self, name, process_info):
        with open(self._process_info_path(name), "w") as process_info_file:
            data = {
                "pid": process_info.pid,
                "startTime": process_info.start_time
            }
            json.dump(data, process_info_file)
    
    def read_process_info(self, name):
        with open(self._process_info_path(name)) as process_info_file:
            properties = json.load(process_info_file)
            return ProcessInfo(properties["pid"], properties["startTime"])
    
    def output_path(self, name):
        return os.path.join(self._run_dir, "{0}.output".format(name))

    def _names_path(self):
        return os.path.join(self._run_dir, "names")
        
    def _process_info_path(self, name):
        return os.path.join(self._run_dir, "{0}.process-info".format(name))



import spur
import spur.ssh


_fields = [
    "hostname",
    "port",
    "user",
    "password",
]


def to_dict(ssh_config):
    return dict(
        (field, getattr(ssh_config, field))
        for field in _fields
        if hasattr(ssh_config, field)
    )
    
    
def from_dict(ssh_dict):
    kwargs = dict(
        (key, value)
        for key, value in ssh_dict.iteritems()
        if key in _fields
    )
    return SshConfig(**kwargs)


class SshConfig(object):
    def __init__(self, hostname, port, user, password):
        self.hostname = hostname
        self.port = port
        self.user = user
        self.password = password

    def shell(self):
        return spur.SshShell(
            hostname=self.hostname,
            port=self.port,
            username=self.user,
            password=self.password,
            missing_host_key=spur.ssh.MissingHostKey.accept,
        )



import time
import sys

    
def wait_until(predicate, *args, **kwargs):
    error_message = kwargs.pop("error_message", "Timeout expired")
    
    def on_error(result):
        raise RuntimeError(error_message)
    
    def condition():
        result = predicate()
        return result, result
    
    return _wait(condition, *args, on_error=on_error, **kwargs)


def wait_until_not(predicate, *args, **kwargs):
    return wait_until(lambda: not predicate(), *args, **kwargs)


def wait_until_successful(predicate, errors, *args, **kwargs):
    def try_predicate():
        try:
            return True, predicate()
        except errors:
            return False, sys.exc_info()
    
    def on_error(last_exception):
        raise last_exception[0], last_exception[1], last_exception[2]
    
    return _wait(try_predicate, *args, on_error=on_error, **kwargs)


def _wait(condition, timeout, wait_time=None, on_error=None):
    start_time = time.time()
    while True:
        finished, result = condition()
        if finished:
            return result
            
        if time.time() - start_time > timeout:
            return on_error(result)
            
        time.sleep(wait_time)



import re
import uuid


def dict_to_obj(dict_kwargs, cls):
    cls_kwargs = dict(
        (_from_camel_case(key), value)
        for key, value in dict_kwargs.iteritems()
    )
    
    return cls(**cls_kwargs)


def obj_to_dict(obj):
    return dict(
        (_to_camel_case(key), getattr(obj, key))
        for key in getattr(obj, _fields_attr)
    )


def _from_camel_case(string):
    # http://stackoverflow.com/questions/1175208
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def _to_camel_case(value):
    # http://stackoverflow.com/questions/4303492
    def camelcase(): 
        yield lambda s: s.lower()
        while True:
            yield lambda s: s.capitalize()

    c = camelcase()
    return "".join(c.next()(x) if x else '_' for x in value.split("_"))


_fields_attr = str(uuid.uuid4())


def data_class(name, fields):
    def __init__(self, *args, **kwargs):
        for field_index, field_name in enumerate(fields):
            if field_index < len(args):
                setattr(self, field_name, args[field_index])
            elif field_name in kwargs:
                setattr(self, field_name, kwargs.pop(field_name))
            else:
                raise TypeError("Missing argument: {0}".format(field_name))
                
        for field_name in kwargs.iterkeys():
            raise TypeError("{0}.__init__ does not take keyword argument {1}".format(name, field_name))
    
    def __eq__(self, other):
        if isinstance(other, new_type):
            return all(
                getattr(self, field_name) == getattr(other, field_name)
                for field_name in fields
            )
        else:
            return NotImplemented
        
    def __ne__(self, other):
        return not (self == other)
        
    def __repr__(self):
        values = (getattr(self, field_name) for field_name in fields)
        return "{0}({1})".format(name, ", ".join(values))
        
    def __str__(self):
        return repr(self)
    
    properties = {
        "__init__": __init__,
        "__eq__": __eq__,
        "__ne__": __ne__,
        "__repr__": __repr__,
        "__str__": __str__,
        _fields_attr: fields,
    }
    
    new_type = type(name, (object,), properties)
    return new_type



import threading
import json
import functools
import SocketServer
from wsgiref.simple_server import make_server, WSGIServer

from pyramid.config import Configurator
from pyramid.response import Response

from . import dictobj
from .request import MachineRequest


_default_timeout = 60 * 60


def start_server(port, provider):
    def http_post(func):
        return view(func, "POST")
        
    def http_get(func):
        return view(func, "GET")
    
    
    def view(func, http_method):
        @functools.wraps(func)
        def respond(request):
            if request.method == http_method:
                # TODO: check some credentials
                status_code, result = func(request.json_body, **request.matchdict)
            else:
                status_code, result = 405, "{0} required".format(http_method)
            return Response(
                json.dumps(result),
                status_code=status_code,
                content_type="application/json"
            )
                
        return respond
    
    def success(result):
        return 200, result
        
    def not_found(result):
        return 404, result
        
    def machines(request):
        if request.method == "POST":
            return start(request)
        else:
            return running_machines(request)
    
    @http_post
    def start(body):
        machine_request = dictobj.dict_to_obj(body, MachineRequest)
        machine = provider.start(machine_request)
        return success(_describe_machine(machine))
            
    @http_get
    def running_machines(post):
        machines = provider.list_running_machines()
        return success(map(_describe_machine, machines))
        
    @http_get
    def running_machine(post, identifier):
        machine = provider.find_running_machine(identifier)
        if machine is None:
            return not_found(None)
        else:
            return success(_describe_machine(machine))
    
    @http_get
    def is_running(post, identifier):
        machine = provider.find_running_machine(identifier)
        is_running = machine is not None
        return success({"isRunning": is_running})
    
    @http_get
    def public_port(post, identifier):
        machine = provider.find_running_machine(identifier)
        if machine is None:
            return not_found(None)
        else:
            guest_port = int(post.get("guest-port"))
            public_port = machine.public_port(guest_port)
            return success({"port": public_port})
        
    @http_post
    def restart(post, identifier):
        machine = provider.find_running_machine(identifier)
        if machine is None:
            return not_found(None)
        else:
            machine.restart()
            return success({"status": "OK"})
        
    @http_post
    def destroy(post, identifier):
        machine = provider.find_running_machine(identifier)
        if machine is not None:
            machine.destroy()
        return success({"status": "OK"})
    
    def _describe_machine(machine):
        return {
            "identifier": machine.identifier,
            "name": machine.name,
            "imageName": machine.image_name,
            "hostname": machine.hostname(),
            "users": map(dictobj.obj_to_dict, machine.users()),
        }
    
    config = Configurator()
    
    config.add_route('machines', '/machines')
    config.add_view(machines, route_name='machines')
    
    def add_machine_route(path, view):
        name = path.replace("-", "_")
        path_suffix = "/" + path if path else ""
        config.add_route(name, '/machines/{identifier}' + path_suffix)
        config.add_view(view, route_name=name)
    
    add_machine_route("", running_machine)
    add_machine_route("is-running", is_running)
    add_machine_route("public-port", public_port)
    add_machine_route("restart", restart)
    add_machine_route("destroy", destroy)
    
    app = config.make_wsgi_app()
    
    server = make_server('0.0.0.0', port, app, ThreadedWSGIServer)
    server_thread = threading.Thread(target=server.serve_forever)
    server_thread.start()
    
    return Server(server, server_thread, provider)


class Server(object):
    def __init__(self, server, thread, provider):
        self._server = server
        self._thread = thread
        self._provider = provider
    
    def cron(self):
        if hasattr(self._provider, "cron"):
            self._provider.cron()
    
    def __enter__(self):
        return self
        
    def __exit__(self, *args):
        self._server.shutdown()
        self._thread.join()

class ThreadedWSGIServer(SocketServer.ThreadingMixIn, WSGIServer):
     pass 



from . import dictobj


def request_machine(name, image_name, public_ports=None, timeout=None):
    if public_ports is None:
        public_ports = []
    return MachineRequest(name, image_name, public_ports, timeout)


MachineRequest = dictobj.data_class(
    "MachineRequest",
    ["name", "image_name", "public_ports", "timeout"]
)



import os
import uuid

import spur.ssh

from .sshconfig import SshConfig
from . import wait


class MachineWrapper(object):
    _delegates = [
        "identifier",
        "name",
        "image_name",
        "hostname",
        "is_running",
        "public_port",
        "destroy",
        "users",
    ]
    
    def __init__(self, machine):
        self._machine = machine
        for delegate in self._delegates:
            setattr(self, delegate, getattr(self._machine, delegate))
    
    def __enter__(self):
        return self
        
    def __exit__(self, *args):
        self.destroy()
    
    def root_shell(self):
        root_username = next(
            user.username
            for user in self.users()
            if user.is_root
        )
        return self.shell(root_username)
        
    def shell(self, *args, **kwargs):
        config = self.ssh_config(*args, **kwargs)
        return config.shell()
        
    def ssh_config(self, username=None):
        user = self._find_user(username)
        return SshConfig(
            hostname=self.hostname(),
            port=self.public_port(_GUEST_SSH_PORT),
            user=user.username,
            password=user.password
        )
    
    def restart(self):
        tmp_file = os.path.join("/tmp/", str(uuid.uuid4()))
        with self.root_shell() as root_shell:
            root_shell.run(["touch", tmp_file])
            root_shell.spawn(["reboot"])
            
        def has_restarted():
            try:
                # TODO: automatic reconnection of shell
                with self.root_shell() as root_shell:
                    result = root_shell.run(
                        ["test", "-f", tmp_file],
                        allow_error=True
                    )
                return result.return_code == 1
            except spur.ssh.ConnectionError:
                return False
            
        wait.wait_until(
            has_restarted, timeout=30, wait_time=1,
            error_message="Failed to restart VM"
        )
            
    def _find_user(self, username):
        
        if username is None:
            condition = lambda user: not user.is_root
        else:
            condition = lambda user: user.username == username
            
        return filter(condition, self.users())[0]
        

# TODO: don't assume SSH port of 22
_GUEST_SSH_PORT = 22



#!/usr/bin/env python

import os
from distutils.core import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='peachtree',
    version='0.3.2',
    description='Library for starting and interacting with qemu/kvm',
    long_description=read("README"),
    author='Michael Williamson',
    url='http://github.com/mwilliamson/peachtree',
    scripts=["scripts/peachtree", "scripts/peachtree-server"],
    packages=['peachtree', 'peachtree.qemu'],
    install_requires=[
        "spur>=0.3.0",
        "starboard>=0.1.2",
        "requests>=1.0,<2",
        "pyramid>=1.4b3,<1.5",
        "psutil>=0.6.1,<0.7",
        "futures>=2.1.3",
    ],
)



import os
import uuid
import time
import json
import errno
import random

from concurrent import futures
import spur
import spur.ssh
import starboard

from .. import wait
from ..users import User
from ..machines import MachineWrapper
from .. import processes
from .. import dictobj
from ..request import request_machine, MachineRequest
from . import networkconfig

local_shell = spur.LocalShell()


def qemu_provider(command=None, accel_arg=None, networking=None, data_dir=None):
    if accel_arg is None:
        accel_arg = "kvm:tcg"
    
    if command is None:
        command = _find_qemu_command()
    
    if networking is None:
        networking = UserNetworking()
        
    data_dir = data_dir or _default_data_dir()
    images = Images(data_dir)
    invoker = QemuInvoker(command, accel_arg)
    statuses = Statuses(os.path.join(data_dir, "status"))
    return Provider(invoker, images, networking, statuses)


def _find_qemu_command():
    for command in ["qemu", "kvm"]:
        if local_shell.run(["which", command], allow_error=True).return_code == 0:
            return command
    raise RuntimeError("Could not find qemu")


class Provider(object):
    def __init__(self, invoker, images, networking, statuses):
        self._invoker = invoker
        self._images = images
        self._networking = networking
        self._statuses = statuses
    
    def start(self, *args, **kwargs):
        if len(args) == 1 and not kwargs and isinstance(args[0], MachineRequest):
            request = args[0]
        else:
            request = request_machine(*(["peachtree"] + list(args)), **kwargs)
        network = self._networking.settings_for(request)
        machine = self._start_with_network_settings(request, network)
        
        config = self._guest_network_config_for(machine)
        machine.root_shell().run(self._append_host_command(
            "127.0.0.1", machine.name, config
        ))
        
        return machine
            
    def start_many(self, requests):
        # TODO: assert name of each request is unique
        with futures.ThreadPoolExecutor(max_workers=4) as executor:
            network = self._networking.start_network()
            
            def start(request):
                network_settings = network.settings_for(request)
                return self._start_with_network_settings(request, network_settings)
            
            machines = list(executor.map(start, requests))
            
            addresses = []
            for index, machine in enumerate(machines):
                request = requests[index]
                
                eth1_address = "192.168.0.{0}".format(1 + index)
                netmask = "255.255.255.0"
                with machine.root_shell() as root_shell:
                    config = self._guest_network_config_for(machine)
                    config.configure_internal_interface(
                        root_shell, eth1_address, netmask
                    )
                    addresses.append((request.name, eth1_address))
            
            for machine in machines:
                config = self._guest_network_config_for(machine)
                with machine.root_shell() as root_shell:
                    for hostname, address in addresses:
                        root_shell.run(self._append_host_command(
                            address, hostname, config
                        ))
            
            return MachineSet(machines)
    
    def _guest_network_config_for(self, machine):
        image = self._images.image(machine.image_name)
        os_family = image.operating_system_family
        return networkconfig.network_config(os_family)
    
    def _append_host_command(self, address, hostname, network_config):
         # TODO: properly escape hosts_path
        sh_command = "echo {0} {1} >> '{2}'".format(
            address, hostname, network_config.hosts_path
        )
        return ["sh", "-c", sh_command]

    def _start_with_network_settings(self, request, network):
        image = self._images.image(request.image_name)
        identifier = str(uuid.uuid4())
        
        process_set = processes.start({})
        self._invoker.start_process(image, network, process_set)
        
        status = MachineStatus(
            name=request.name,
            identifier=identifier,
            image_name=request.image_name,
            # TODO: either re-couple network, or find a better way of
            # storing network details
            forwarded_ports=network.forwarded_ports,
            timeout=request.timeout,
            start_time=time.time(),
            process_set_run_dir=process_set.run_dir
        )
        
        self._statuses.write(status)
        machine = _create_machine(image.users, status, self._statuses)
        
        try:
            self._wait_for_ssh(process_set, machine)
            return machine
        except:
            machine.destroy()
            raise
        
    def _wait_for_ssh(self, process_set, machine):
        def attempt_ssh_command():
            if not process_set.all_running():
                process_set.kill_all()
                wait.wait_until_not(process_set.any_running, timeout=1, wait_time=0.1)
                output = process_set.all_output()
                raise RuntimeError("Process died, output:\n{0}".format(output))
            with machine.root_shell() as shell:
                shell.run(["true"])
            
        wait.wait_until_successful(
            attempt_ssh_command,
            errors=(spur.ssh.ConnectionError, ),
            timeout=60,
            wait_time=1
        )
        
    def find_running_machine(self, identifier):
        machine = self._find_machine(identifier)
        if machine is None:
            return None
        elif machine.is_running():
            return machine
        else:
            return None
        
    def _find_machine(self, identifier):
        status = self._statuses.read(identifier)
        if status is None:
            return None
        else:
            return self._machine_from_status(status)
        
    def _machine_from_status(self, status):
        image = self._images.image(status.image_name)
        return _create_machine(image.users, status, self._statuses)
    
    def list_running_machines(self):
        statuses = self._statuses.read_all()
        machines = map(self._machine_from_status, statuses)
        return [machine for machine in machines if machine is not None]
                
    def cron(self):
        self._stop_machines_past_timeout()
        self._clean_statuses()
        
    def _stop_machines_past_timeout(self):
        statuses = self._statuses.read_all()
        for status in statuses:
            if status.timeout is not None:
                running_time = time.time() - status.start_time
                if running_time > status.timeout:
                    self._machine_from_status(status).destroy()
    
    def _clean_statuses(self):
        for status in self._statuses.read_all():
            machine = self._machine_from_status(status)
            if not machine.is_running():
                self._statuses.remove(status.identifier)


class QemuInvoker(object):
    def __init__(self, command, accel_arg):
        self._command = command
        self._accel_arg = accel_arg
        
    def start_process(self, image, network, process_set):
        disk_args = []
        for disk in image.disks:
            disk_args += ["-drive", "file={0},if=virtio".format(disk)]
        
        qemu_command = [
            self._command, "-machine", "accel={0}".format(self._accel_arg),
            "-snapshot",
            "-nographic", "-serial", "none",
            "-m", str(image.memory_size),
        ] + disk_args + network.qemu_args()
        process_set.start({"qemu": qemu_command})


class Images(object):
    def __init__(self, data_dir=None):
        self._data_dir = data_dir or _default_data_dir()
    
    def image_path(self, image_name):
        return os.path.join(self._data_dir, "images", image_name)
    
    def image(self, image_name):
        image_dir = self.image_path(image_name)
        with open(os.path.join(image_dir, "image.json")) as description_file:
            description = json.load(description_file)
        relative_disks = description["disks"]
        disks = [
            os.path.abspath(os.path.join(image_dir, relative_disk))
            for relative_disk in relative_disks
        ]
        memory_size = description.get("memory", 512)
        
        users_json = description.get("users", None)
        if users_json is None:
            password = "password1"
            users = [
                User("qemu-user", password, is_root=False),
                User("root", password, is_root=True),
            ]
        else:
            users = [
                dictobj.dict_to_obj(user_json, User)
                for user_json in users_json
            ]
            
        operating_system_family = description.get("operatingSystemFamily", "linux")
            
        return Image(disks, memory_size, users, operating_system_family)


class Image(object):
    def __init__(self, disks, memory_size, users, operating_system_family):
        self.disks = disks
        self.memory_size = memory_size
        self.users = users
        self.operating_system_family = operating_system_family
    
    
def _default_data_dir():
    xdg_data_home = os.environ.get("XDG_DATA_HOME", os.path.expanduser("~/.local/share"))
    return os.path.join(xdg_data_home, "peachtree-0.3/qemu")
    

MachineStatus = dictobj.data_class("MachineStatus",
    [
        "identifier",
        "name",
        "image_name",
        "forwarded_ports",
        "start_time",
        "timeout",
        "process_set_run_dir",
    ]
)


class MachineSet(object):
    def __init__(self, machines):
        self._machines = machines
        
    def __enter__(self):
        return self
        
    def __exit__(self, *args):
        for machine in self._machines:
            machine.destroy()
            
    def __getitem__(self, key):
        if isinstance(key, basestring):
            return next(
                machine
                for machine in self._machines
                if machine.name == key
            )
        elif isinstance(key, (int, long)):
            return self._machines[key]
        else:
            raise TypeError("Expected key to be string or integer")


class Statuses(object):
    def __init__(self, status_dir):
        self._status_dir = status_dir
        
    def remove(self, identifier):
        try:
            os.remove(self._status_path(identifier))
        except OSError as error:
            # ENOENT: Machine has been shut down in the interim, so ignore
            if error.errno != errno.ENOENT:
                raise
    
    def write(self, status):
        status_json = dictobj.obj_to_dict(status)
        self._write_json(status.identifier, status_json)
    
    def read(self, identifier):
        try:
            status_dict = self._read_json(identifier)
        except IOError as error:
            # ENOENT: Machine has been shut down in the interim, so ignore
            if error.errno == errno.ENOENT:
                return None
            else:
                raise
        
        status_dict["forwardedPorts"] = dict(
            (int(guest_port), host_port)
            for guest_port, host_port
            in status_dict["forwardedPorts"].iteritems()
        )
        status = dictobj.dict_to_obj(status_dict, MachineStatus)
        return status
                        
    def read_all(self):
        if not os.path.exists(self._status_dir):
            return []
        identifiers = os.listdir(self._status_dir)
        statuses = map(self.read, identifiers)
        return filter(lambda status: status is not None, statuses)
    
    def _status_path(self, identifier):
        return os.path.join(self._status_dir, identifier)

    def _read_json(self, identifier):
        status_file_path = self._status_path(identifier)
        with open(status_file_path) as status_file:
            return json.load(status_file)
    
    def _write_json(self, identifier, data):
        status_path = self._status_path(identifier)
        
        if not os.path.exists(os.path.dirname(status_path)):
            os.makedirs(os.path.dirname(status_path))
            
        with open(status_path, "w") as status_file:
            json.dump(data, status_file)


def _create_machine(*args, **kwargs):
    machine = QemuMachine(*args, **kwargs)
    return MachineWrapper(machine)


class QemuMachine(object):
    def __init__(self, users, status, statuses):
        self._users = users
        self.name = status.name
        self.image_name = status.image_name
        self.identifier = status.identifier
        self._process_set = processes.from_dir(status.process_set_run_dir)
        self._forwarded_ports = status.forwarded_ports
        self._statuses = statuses
    
    def is_running(self):
        return self._process_set.all_running()
    
    def destroy(self):
        self._process_set.kill_all()
        
        wait.wait_until_not(
            self._process_set.any_running, timeout=10, wait_time=0.1,
            error_message="Failed to kill VM {0}".format(self.identifier)
        )
        
        self._statuses.remove(self.identifier)
        
    def public_port(self, guest_port):
        return self._forwarded_ports.get(guest_port, None)
        
    def hostname(self):
        return starboard.find_local_hostname()
        
    def users(self):
        return self._users


_GUEST_SSH_PORT = 22


class UserNetworking(object):
    def settings_for(self, request):
        forwarded_ports = _generate_forwarded_ports(request.public_ports)
        return UserNetworkSettings(forwarded_ports, [])
        
    def start_network(self):
        port = starboard.find_local_free_udp_port()
        return UserNetwork(port)
        
        
class UserNetwork(object):
    def __init__(self, port):
        self._port = port
        
    def settings_for(self, request):
        forwarded_ports = _generate_forwarded_ports(request.public_ports)
        socket_args = _generate_network_args(
            "guest-net-socket",
            "socket,mcast=230.0.0.1:{0},localaddr=127.0.0.1".format(self._port),
        )
        return UserNetworkSettings(forwarded_ports, socket_args)


class UserNetworkSettings(object):
    def __init__(self, forwarded_ports, extra_args):
        self.forwarded_ports = forwarded_ports
        self._extra_args = extra_args
        
    def qemu_args(self):
        kvm_forward_ports = [
            "hostfwd=tcp::{0}-:{1}".format(host_port, guest_port)
            for guest_port, host_port
            in self.forwarded_ports.iteritems()
        ]
        return _generate_network_args(
            "guest-net-user",
            "user,{0}".format(",".join(kvm_forward_ports))
        ) + self._extra_args


def _generate_network_args(name, netdev):
    mac_address_parts = [0x06] + [random.randint(0x00, 0xff) for i in range(5)]
    mac_address = ":".join("{0:02x}".format(part) for part in mac_address_parts)
    return [
        "-netdev", "{0},id={1}".format(netdev, name),
        "-device", "virtio-net-pci,netdev={0},mac={1}".format(name, mac_address),
    ]

    
def _generate_forwarded_ports(public_ports):
    public_ports = set([_GUEST_SSH_PORT] + public_ports)
    host_ports = starboard.find_local_free_tcp_ports(len(public_ports))
    return dict(zip(public_ports, host_ports))



import re

from .. import wait


def network_config(operating_system_family):
    configs = {
        "linux": LinuxNetworkConfig(),
        "windows": WindowsNetworkConfig()
    }
    return configs[operating_system_family]


class LinuxNetworkConfig(object):
    hosts_path = "/etc/hosts"
    
    def configure_internal_interface(self, root_shell, ip_address, netmask):
        root_shell.run(["ifconfig", "eth1", ip_address, "netmask", netmask])
        
        
class WindowsNetworkConfig(object):
    hosts_path = r"C:\Windows\System32\drivers\etc\hosts"
    
    def configure_internal_interface(self, root_shell, ip_address, netmask):
        internal_interface_name = wait.wait_until(
            lambda: self._find_internal_interface_name(root_shell),
            timeout=30, wait_time=0.5
        )
        root_shell.run([
            "netsh", "interface", "ip", "set", "address",
            internal_interface_name, "static", ip_address, netmask
        ])
        
    def _find_internal_interface_name(self, root_shell):
        # The internal network has no DHCP server
        for interface_name in self._find_interface_names(root_shell):
            config_result = root_shell.run([
                "netsh", "interface", "ip", "show", "config",
                "name={0}".format(interface_name)
            ])
            dhcp_search_regex = r"DNS servers configured through DHCP:\s+None"
            if re.search(dhcp_search_regex, config_result.output):
                return interface_name
        
    def _find_interface_names(self, root_shell):
        field_separator_regex = re.compile("\s{2,}")
        show_interfaces_output = root_shell.run([
            "netsh", "interface", "ip", "show", "interface"
        ]).output
        lines = [
            line.strip()
            for line in show_interfaces_output.split("\n")
            if re.search(r"[^\s\-]", line)
        ]
        headers = field_separator_regex.split(lines[0])
        name_column_index = headers.index("Name")
        
        interface_names = [
            field_separator_regex.split(line)[name_column_index]
            for line in lines[1:] # Skip headers
        ]
        
        return filter(
            lambda name: "loopback" not in name.lower(),
            interface_names
        )
